from __future__ import annotations

import os
import sys
from typing import Any, Callable

from hatchling.builders.config import BuilderConfig
from hatchling.builders.plugin.interface import BuilderInterface


class AppBuilderConfig(BuilderConfig):
    SUPPORTED_VERSIONS = ('3.11', '3.10', '3.9', '3.8', '3.7')

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.__scripts: list[str] | None = None
        self.__python_version: str | None = None
        self.__pyapp_version: str | None = None

    @property
    def scripts(self) -> list[str]:
        if self.__scripts is None:
            known_scripts = self.builder.metadata.core.scripts
            scripts = self.target_config.get('scripts', [])

            if not isinstance(scripts, list):
                message = f'Field `tool.hatch.build.targets.{self.plugin_name}.scripts` must be an array'
                raise TypeError(message)

            for i, script in enumerate(scripts, 1):
                if not isinstance(script, str):
                    message = (
                        f'Script #{i} of field `tool.hatch.build.targets.{self.plugin_name}.scripts` must be a string'
                    )
                    raise TypeError(message)
                elif script not in known_scripts:
                    message = f'Unknown script in field `tool.hatch.build.targets.{self.plugin_name}.scripts`: {script}'
                    raise ValueError(message)

            self.__scripts = sorted(set(scripts)) if scripts else list(known_scripts)

        return self.__scripts

    @property
    def python_version(self) -> str:
        if self.__python_version is None:
            python_version = self.target_config.get('python-version', '')

            if not isinstance(python_version, str):
                message = f'Field `tool.hatch.build.targets.{self.plugin_name}.python-version` must be a string'
                raise TypeError(message)

            if not python_version and 'PYAPP_DISTRIBUTION_SOURCE' not in os.environ:
                for supported_version in self.SUPPORTED_VERSIONS:
                    if self.builder.metadata.core.python_constraint.contains(supported_version):
                        python_version = supported_version
                        break
                else:
                    message = 'Field `project.requires-python` is incompatible with the known distributions'
                    raise ValueError(message)

            self.__python_version = python_version

        return self.__python_version

    @property
    def pyapp_version(self) -> str:
        if self.__pyapp_version is None:
            pyapp_version = self.target_config.get('pyapp-version', '')

            if not isinstance(pyapp_version, str):
                message = f'Field `tool.hatch.build.targets.{self.plugin_name}.pyapp-version` must be a string'
                raise TypeError(message)

            self.__pyapp_version = pyapp_version

        return self.__pyapp_version


class AppBuilder(BuilderInterface):
    """
    Build applications
    """

    PLUGIN_NAME = 'app'

    def get_version_api(self) -> dict[str, Callable]:
        return {'bootstrap': self.build_bootstrap}

    def get_default_versions(self) -> list[str]:
        return ['bootstrap']

    def clean(self, directory: str, versions: list[str]) -> None:
        import shutil

        app_dir = os.path.join(directory, 'app')
        if os.path.isdir(app_dir):
            shutil.rmtree(app_dir)

    def build_bootstrap(self, directory: str, **build_data: Any) -> str:
        import shutil
        import tempfile

        cargo_path = os.environ.get('CARGO', '')
        if not cargo_path:
            if not shutil.which('cargo'):
                message = 'Executable `cargo` could not be found on PATH'
                raise OSError(message)

            cargo_path = 'cargo'

        app_dir = os.path.join(directory, 'app')
        if not os.path.isdir(app_dir):
            os.makedirs(app_dir)

        on_windows = sys.platform == 'win32'
        base_env = dict(os.environ)
        base_env['PYAPP_PROJECT_NAME'] = self.metadata.name
        base_env['PYAPP_PROJECT_VERSION'] = self.metadata.version

        if self.config.python_version:
            base_env['PYAPP_PYTHON_VERSION'] = self.config.python_version

        # https://doc.rust-lang.org/cargo/reference/config.html#buildtarget
        build_target = os.environ.get('CARGO_BUILD_TARGET', '')

        # This will determine whether we install from crates.io or build locally and is currently required for
        # cross compilation: https://github.com/cross-rs/cross/issues/1215
        repo_path = os.environ.get('PYAPP_REPO', '')

        with tempfile.TemporaryDirectory() as temp_dir:
            exe_name = 'pyapp.exe' if on_windows else 'pyapp'
            if repo_path:
                context_dir = repo_path
                target_dir = os.path.join(temp_dir, 'build')
                if build_target:
                    temp_exe_path = os.path.join(target_dir, build_target, 'release', exe_name)
                else:
                    temp_exe_path = os.path.join(target_dir, 'release', exe_name)
                install_command = [cargo_path, 'build', '--release', '--target-dir', target_dir]
            else:
                context_dir = temp_dir
                temp_exe_path = os.path.join(temp_dir, 'bin', exe_name)
                install_command = [cargo_path, 'install', 'pyapp', '--force', '--root', temp_dir]
                if self.config.pyapp_version:
                    install_command.extend(['--version', self.config.pyapp_version])

            if self.config.scripts:
                for script in self.config.scripts:
                    env = dict(base_env)
                    env['PYAPP_EXEC_SPEC'] = self.metadata.core.scripts[script]

                    self.cargo_build(install_command, cwd=context_dir, env=env)

                    exe_stem = f'{script}-{build_target}' if build_target else script
                    exe_path = os.path.join(app_dir, f'{exe_stem}.exe' if on_windows else exe_stem)
                    shutil.move(temp_exe_path, exe_path)
            else:
                self.cargo_build(install_command, cwd=context_dir, env=base_env)

                exe_stem = f'{self.metadata.name}-{build_target}' if build_target else self.metadata.name
                exe_path = os.path.join(app_dir, f'{exe_stem}.exe' if on_windows else exe_stem)
                shutil.move(temp_exe_path, exe_path)

        return app_dir

    def cargo_build(self, *args: Any, **kwargs: Any) -> None:
        import subprocess

        if not self.app.verbosity:
            kwargs['stdout'] = subprocess.PIPE
            kwargs['stderr'] = subprocess.STDOUT

        process = subprocess.run(*args, **kwargs)
        if process.returncode:
            message = f'Compilation of failed (code {process.returncode})'
            if not self.app.verbosity:
                message = f'{process.stdout.decode("utf-8")}\n{message}'

            raise OSError(message)

    @classmethod
    def get_config_class(cls) -> type[AppBuilderConfig]:
        return AppBuilderConfig



import errno
import json
import os
import platform
import shutil
import stat
import subprocess
import sys
import tempfile
from contextlib import contextmanager
from zipfile import ZipFile

import requests
import tomli
from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet
from virtualenv import cli_run

HERE = os.path.dirname(os.path.abspath(__file__))
ON_WINDOWS = platform.system() == 'Windows'


def handle_remove_readonly(func, path, exc):  # no cov
    # PermissionError: [WinError 5] Access is denied: '...\\.git\\...'
    if func in (os.rmdir, os.remove, os.unlink) and exc[1].errno == errno.EACCES:
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)  # noqa: S103
        func(path)
    else:
        raise


class EnvVars(dict):
    def __init__(self, env_vars=None, ignore=None):
        super().__init__(os.environ)
        self.old_env = dict(self)

        if env_vars is not None:
            self.update(env_vars)

        if ignore is not None:
            for env_var in ignore:
                self.pop(env_var, None)

    def __enter__(self):
        os.environ.clear()
        os.environ.update(self)

    def __exit__(self, exc_type, exc_value, traceback):
        os.environ.clear()
        os.environ.update(self.old_env)


def python_version_supported(project_config):
    requires_python = project_config['project'].get('requires-python', '')
    if requires_python:
        python_constraint = SpecifierSet(requires_python)
        if not python_constraint.contains(str('.'.join(map(str, sys.version_info[:2])))):
            return False

    return True


def download_file(url, file_name):
    response = requests.get(url, stream=True, timeout=20)
    with open(file_name, 'wb') as f:
        for chunk in response.iter_content(16384):
            f.write(chunk)


@contextmanager
def temp_dir():
    d = tempfile.mkdtemp()

    try:
        d = os.path.realpath(d)
        yield d
    finally:
        shutil.rmtree(d, ignore_errors=False, onerror=handle_remove_readonly)


def get_venv_exe_dir(venv_dir):
    exe_dir = os.path.join(venv_dir, 'Scripts' if ON_WINDOWS else 'bin')
    if os.path.isdir(exe_dir):
        return exe_dir
    # PyPy
    elif ON_WINDOWS:
        exe_dir = os.path.join(venv_dir, 'bin')
        if os.path.isdir(exe_dir):
            return exe_dir
        else:
            message = f'Unable to locate executables directory within: {venv_dir}'
            raise OSError(message)
    # Debian
    elif os.path.isdir(os.path.join(venv_dir, 'local')):
        exe_dir = os.path.join(venv_dir, 'local', 'bin')
        if os.path.isdir(exe_dir):
            return exe_dir
        else:
            message = f'Unable to locate executables directory within: {venv_dir}'
            raise OSError(message)
    else:
        message = f'Unable to locate executables directory within: {venv_dir}'
        raise OSError(message)


def main():
    original_backend_path = os.path.dirname(os.path.dirname(HERE))
    with temp_dir() as links_dir, temp_dir() as build_dir:
        print('<<<<< Copying backend >>>>>')
        backend_path = os.path.join(build_dir, 'backend')
        shutil.copytree(original_backend_path, backend_path)

        # Increment the minor version
        version_file = os.path.join(backend_path, 'src', 'hatchling', '__about__.py')
        with open(version_file) as f:
            lines = f.readlines()

        for i, line in enumerate(lines):
            if line.startswith('__version__'):
                version = line.strip().split(' = ')[1].strip('\'"')
                version_parts = version.split('.')
                version_parts[1] = str(int(version_parts[1]) + 1)
                lines[i] = line.replace(version, '.'.join(version_parts))
                break
        else:
            message = 'No version found'
            raise ValueError(message)

        with open(version_file, 'w') as f:
            f.writelines(lines)

        print('<<<<< Building backend >>>>>')
        subprocess.check_call([sys.executable, '-m', 'build', '--wheel', '-o', links_dir, backend_path])
        subprocess.check_call(
            [
                sys.executable,
                '-m',
                'pip',
                'download',
                '-q',
                '--disable-pip-version-check',
                '--no-python-version-warning',
                '-d',
                links_dir,
                os.path.join(links_dir, os.listdir(links_dir)[0]),
            ]
        )

        constraints = []
        constraints_file = os.path.join(build_dir, 'constraints.txt')
        with open(constraints_file, 'w') as f:
            f.write('\n'.join(constraints))

        for project in os.listdir(HERE):
            project_dir = os.path.join(HERE, project)
            if not os.path.isdir(project_dir):
                continue

            print(f'<<<<< Project: {project} >>>>>')
            project_config = {}
            potential_project_file = os.path.join(project_dir, 'pyproject.toml')

            # Not yet ported
            if os.path.isfile(potential_project_file):
                with open(potential_project_file) as f:
                    project_config.update(tomli.loads(f.read()))

                if not python_version_supported(project_config):
                    print('--> Unsupported version of Python, skipping')
                    continue

            with open(os.path.join(project_dir, 'data.json')) as f:
                test_data = json.loads(f.read())

            with temp_dir() as d:
                if 'repo_url' in test_data:
                    print('--> Cloning repository')
                    repo_dir = os.path.join(d, 'repo')
                    subprocess.check_call(['git', 'clone', '-q', '--depth', '1', test_data['repo_url'], repo_dir])
                else:
                    archive_name = f'{project}.zip'
                    archive_path = os.path.join(d, archive_name)

                    print('--> Downloading archive')
                    download_file(test_data['archive_url'], archive_path)
                    with ZipFile(archive_path) as zip_file:
                        zip_file.extractall(d)

                    entries = os.listdir(d)
                    entries.remove(archive_name)
                    repo_dir = os.path.join(d, entries[0])

                project_file = os.path.join(repo_dir, 'pyproject.toml')
                if project_config:
                    shutil.copyfile(potential_project_file, project_file)
                else:
                    if not os.path.isfile(project_file):
                        sys.exit('--> Missing file: pyproject.toml')

                    with open(project_file) as f:
                        project_config.update(tomli.loads(f.read()))

                    for requirement in project_config.get('build-system', {}).get('requires', []):
                        if Requirement(requirement).name == 'hatchling':
                            break
                    else:
                        sys.exit('--> Field `build-system.requires` must specify `hatchling` as a requirement')

                    if not python_version_supported(project_config):
                        print('--> Unsupported version of Python, skipping')
                        continue

                for file_name in ('MANIFEST.in', 'setup.cfg', 'setup.py'):
                    possible_path = os.path.join(repo_dir, file_name)
                    if os.path.isfile(possible_path):
                        os.remove(possible_path)

                venv_dir = os.path.join(d, '.venv')
                print('--> Creating virtual environment')
                cli_run([venv_dir, '--no-download', '--no-periodic-update'])

                env_vars = dict(test_data.get('env_vars', {}))
                env_vars['VIRTUAL_ENV'] = venv_dir
                env_vars['PATH'] = f'{get_venv_exe_dir(venv_dir)}{os.pathsep}{os.environ["PATH"]}'
                env_vars['PIP_CONSTRAINT'] = constraints_file
                with EnvVars(env_vars, ignore=('__PYVENV_LAUNCHER__', 'PYTHONHOME')):
                    print('--> Installing project')
                    subprocess.check_call(
                        [
                            shutil.which('pip'),
                            'install',
                            '-q',
                            '--disable-pip-version-check',
                            '--no-python-version-warning',
                            '--find-links',
                            links_dir,
                            '--no-deps',
                            repo_dir,
                        ]
                    )

                    print('--> Installing dependencies')
                    subprocess.check_call(
                        [
                            shutil.which('pip'),
                            'install',
                            '-q',
                            '--disable-pip-version-check',
                            '--no-python-version-warning',
                            repo_dir,
                        ]
                    )

                    print('--> Testing package')
                    for statement in test_data['statements']:
                        subprocess.check_call([shutil.which('python'), '-c', statement])

                    scripts = project_config['project'].get('scripts', {})
                    if scripts:
                        print('--> Testing scripts')
                        for script in scripts:
                            if not shutil.which(script):
                                sys.exit(f'--> Could not locate script: {script}')

                    print('--> Success!')


if __name__ == '__main__':
    main()
